@@ -21,16 +21,23 @@
 
 #include <exception>
 #include <iostream>
+#include <list>
+#include <iterator>
 #include <cxxtools/httprequest.h>
 #include <cxxtools/httpreply.h>
 #include <cxxtools/arg.h>
+#include <cxxtools/base64stream.h>
 
 int main(int argc, char* argv[])
 {
   try
   {
     cxxtools::Arg<bool> raw(argc, argv, 'r');
+    cxxtools::Arg<bool> headers(argc, argv, 'h');
+    cxxtools::Arg<bool> returncode(argc, argv, 'c');
+    cxxtools::Arg<bool> body(argc, argv, 'b');
     cxxtools::Arg<bool> debug(argc, argv, 'd');
+    cxxtools::Arg<std::string> user(argc, argv, 'u');
 
     cxxtools::Arg<std::string> url("http://127.0.0.1/");
     url.set(argc, argv);
@@ -38,6 +45,11 @@ int main(int argc, char* argv[])
     // set parameters
     cxxtools::HttpRequest request(url);
 
+    std::string::size_type p = user.getValue().find('/');
+    if (p != std::string::npos)
+      request.setAuth(user.getValue().substr(0, p),
+                      user.getValue().substr(p + 1));
+
     if (debug)
     {
       std::cout << "host: " << request.getHost() << "\n"
@@ -56,8 +68,25 @@ int main(int argc, char* argv[])
     {
       // let httpreply parse the http-headers and read only body
       cxxtools::HttpReply reply(request);
-      std::cout << "return-code " << reply.getReturnCode() << '\n'
-                << reply.rdbuf() << std::flush;
+
+      if (headers)
+      {
+        typedef std::list<std::string> headerlistType;
+        headerlistType headerlist;
+        reply.getHeaders(std::back_inserter(headerlist));
+        for (headerlistType::iterator it = headerlist.begin();
+             it != headerlist.end(); ++it)
+          std::cout << *it << ":\t" << reply.getHeader(*it) << '\n';
+        std::cout << std::endl;
+      }
+
+      if (returncode)
+        std::cout << "return-code " << reply.getReturnCode() << '\n';
+
+      if (!headers || body)
+        std::cout << reply.rdbuf();
+
+      std::cout.flush();
     }
   }
   catch (const std::exception& e)
