@@ -25,7 +25,9 @@
 #include <cxxtools/udpstream.h>
 #include <cxxtools/tee.h>
 #include <cxxtools/streamcounter.h>
+#include <cxxtools/pipestream.h>
 #include <list>
+#include <vector>
 #include <algorithm>
 #include <fstream>
 #include <sstream>
@@ -33,12 +35,51 @@
 #include <time.h>
 #include <sys/time.h>
 #include <sys/stat.h>
+#include <sys/types.h>
+#include <sys/wait.h>
 #include <unistd.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <cctype>
+#include <pwd.h>
+#include <grp.h>
+#include <config.h>
+#ifdef HAVE_SCHED_YIELD
+#include <sched.h>
+#endif
+
+log_define("cxxtools.log")
 
 namespace cxxtools
 {
+  namespace
+  {
+    void setUserAndGroup(struct passwd* pw, struct group* gr)
+    {
+      int gret, pret;
+      if (gr)
+        gret = ::setgid(gr->gr_gid);
+      if (pw)
+        pret = ::setuid(pw->pw_uid);
+
+      if (gr)
+      {
+        if (gret == 0)
+          log_debug("logging group changed to " << gr->gr_name << '(' << gr->gr_gid << ')');
+        else
+          log_warn("error changing logging group to " << gr->gr_name << '(' << gr->gr_gid << ')');
+      }
+
+      if (pw)
+      {
+        if (pret == 0)
+          log_debug("logging user changed to " << pw->pw_name << '(' << pw->pw_uid << ')');
+        else
+          log_warn("error changing logging user to " << pw->pw_name << '(' << pw->pw_uid << ')');
+      }
+    }
+  }
+
   class LoggerImpl : public Logger
   {
       static std::string fname;
@@ -49,6 +90,7 @@ namespace cxxtools
       static net::UdpOStream udpmessage;
       static unsigned maxfilesize;
       static unsigned maxbackupindex;
+      static Pipestream* pipe;
 
       static std::string mkfilename(unsigned idx);
 
@@ -87,11 +129,14 @@ namespace cxxtools
       static void setMaxFileSize(unsigned size)   { maxfilesize = size; }
       static void setMaxBackupIndex(unsigned idx) { maxbackupindex = idx; }
       static void setLoghost(const std::string& host, unsigned short int port);
+      static void runLoggerProcess(const std::string& user, const std::string& group);
   };
 
   std::ostream& LoggerImpl::getAppender()
   {
-    if (!fname.empty())
+    if (pipe)
+      return *pipe;
+    else if (!fname.empty())
     {
       if (!outfile.is_open())
       {
@@ -168,6 +213,7 @@ namespace cxxtools
   unsigned LoggerImpl::maxbackupindex = 0;
   LoggerImpl::FlusherThread* LoggerImpl::flusherThread = 0;
   bool LoggerImpl::flusherThreadStarted = false;
+  Pipestream* LoggerImpl::pipe = 0;
 
   struct timespec LoggerImpl::FlusherThread::flushDelay;
 
@@ -193,6 +239,73 @@ namespace cxxtools
   {
     loghost.connect(host.c_str(), port, true);
   }
+  
+  void LoggerImpl::runLoggerProcess(const std::string& user, const std::string& group)
+  {
+    struct passwd * pw = 0;
+    if (!user.empty())
+    {
+      pw = getpwnam(user.c_str());
+      if (pw == 0)
+        throw std::runtime_error("unknown user \"" + user + "\" in logging configuration");
+    }
+
+    struct group * gr = 0;
+    if (!group.empty())
+    {
+      gr = getgrnam(group.c_str());
+      if (gr == 0)
+        throw std::runtime_error("unknown group \"" + group + "\" in logging configuration");
+    }
+
+    pipe = new Pipestream();
+    pid_t pid = ::fork();
+    if (pid < 0)
+      throw SysError("fork");
+    if (pid == 0)
+    {
+      // 1st child
+      pipe->closeWriteFd();
+      pid = ::fork();
+      if (pid < 0)
+        exit(-1);
+      if (pid)
+        exit(0);  // exit middle process
+
+      // 2nd child
+
+      std::streambuf* in = pipe->rdbuf();
+      // set global pipe pointer to 0, so that getAppender do not return
+      // that pipe, but skips to the next appender
+      pipe = 0;
+
+      setUserAndGroup(pw, gr);
+
+      counter.resetCount(outfile.tellp());
+
+
+      log_debug("logger process initialized");
+      char ich;
+      std::ostream& out = getAppender();
+      while ((ich = in->snextc()) != std::ios::traits_type::eof())
+      {
+        char ch = std::ios::traits_type::to_char_type(ich);
+        out.rdbuf()->sputc(ch);
+        if (ch == '\n')
+          getAppender();
+      }
+      exit(0);
+    }
+    else
+    {
+      // parent
+      pipe->closeReadFd();
+      int status;
+      ::waitpid(pid, &status, 0);
+      if (WEXITSTATUS(status) != 0)
+        throw std::runtime_error("error creating logging process");
+    }
+  }
 
   RWLock Logger::rwmutex;
   Mutex Logger::mutex;
@@ -265,7 +378,7 @@ namespace cxxtools
     }
   }
 
-  Logger* Logger::getLogger(const std::string& category)
+  Logger* Logger::getCategoryLogger(const std::string& category)
   {
     if (!enabled)
       return 0;
@@ -338,7 +451,7 @@ namespace cxxtools
     { return digits[i % 10]; }
   }
 
-  std::ostream& Logger::logentry(const char* level)
+  std::ostream& Logger::logentry(std::ostream& out, const char* level, const std::string& category)
   {
     struct timeval t;
     gettimeofday(&t, 0);
@@ -375,10 +488,6 @@ namespace cxxtools
       date[20] = '\0';
     }
 
-    std::ostream& out = LoggerImpl::getAppender();
-
-    out.clear();
-
     out << date
         << loDigit(t.tv_usec / 100000) << loDigit(t.tv_usec / 10000)
         << loDigit(t.tv_usec / 1000) << loDigit(t.tv_usec / 100)
@@ -390,12 +499,25 @@ namespace cxxtools
     return out;
   }
 
+  std::ostream& Logger::logentry(const char* level)
+  {
+    std::ostream& out = LoggerImpl::getAppender();
+    out.clear();
+    logentry(out, level, category);
+    return out;
+  }
+
   class LogMessage::LogMessageImpl
   {
       std::ostringstream msg;
       Logger* logger;
       const char* level;
 
+      static Mutex queueMutex;
+      typedef std::vector<std::string> MessageQueueType;
+      static MessageQueueType messageQueue;
+      void flushQueue();
+
     public:
       LogMessageImpl(Logger* logger_, const char* level_)
         : logger(logger_),
@@ -403,21 +525,82 @@ namespace cxxtools
           { }
 
       std::ostream& out()     { return msg; }
-      void flush()
-      {
-        MutexLock lock(Logger::mutex);
-
-        std::ostream& out(logger->logentry(level));
-        out << msg.str() << '\n';
-
-        logger->logEnd(out);
-      }
+      void flush();
   };
 
+  Mutex LogMessage::LogMessageImpl::queueMutex;
+  LogMessage::LogMessageImpl::MessageQueueType LogMessage::LogMessageImpl::messageQueue;
+
+  void LogMessage::LogMessageImpl::flush()
+  {
+    if (Logger::mutex.tryLock())
+    {
+      UnlockMonitor<> unlocker(Logger::mutex);
+
+      std::ostream& out(logger->logentry(level));
+      out << msg.str() << '\n';
+
+      MutexLock lock(queueMutex);
+      flushQueue();
+      unlocker.unlock();
+    }
+    else
+    {
+      std::ostringstream logmessage;
+      Logger::logentry(logmessage, level, logger->getCategory());
+      logmessage << msg.str();
+
+#ifdef HAVE_SCHED_YIELD
+      sched_yield();
+#endif
+
+      MutexLock lock(queueMutex);
+      messageQueue.push_back(logmessage.str());
+
+      if (Logger::mutex.tryLock())
+      {
+        UnlockMonitor<> unlocker(Logger::mutex);
+        flushQueue();
+      }
+    }
+  }
+
+  void LogMessage::LogMessageImpl::flushQueue()
+  {
+    std::ostream& out = LoggerImpl::getAppender();
+    if (!messageQueue.empty())
+    {
+      /*
+      static unsigned count = 0;
+      count += messageQueue.size();
+      if (messageQueue.size() == 1)
+        std::cerr << '.';
+      else
+        std::cerr << '[' << messageQueue.size() << '/' << count << ']';
+      */
+      for (MessageQueueType::const_iterator it = messageQueue.begin();
+           it != messageQueue.end(); ++it)
+        out << *it << '\n';
+      messageQueue.clear();
+    }
+    logger->logEnd(out);
+  }
+
   LogMessage::LogMessage(Logger* logger, const char* level)
     : impl(new LogMessageImpl(logger, level))
     { }
 
+  LogMessage::LogMessage(Logger* logger, Logger::log_level_type level)
+    : impl(new LogMessageImpl(logger,
+                              level >= Logger::LOG_LEVEL_TRACE ? "TRACE"
+                            : level >= Logger::LOG_LEVEL_DEBUG ? "DEBUG"
+                            : level >= Logger::LOG_LEVEL_INFO ? "INFO"
+                            : level >= Logger::LOG_LEVEL_WARN ? "WARN"
+                            : level >= Logger::LOG_LEVEL_ERROR ? "ERROR"
+                            : "FATAL"))
+    { }
+
+
   void LogMessage::flush()
   {
     impl->flush();
@@ -516,6 +699,11 @@ void log_init_cxxtools(std::istream& in)
     state_flushdelay0,
     state_flushdelay,
     state_disable,
+    state_logprocess,
+    state_logprocessuser0,
+    state_logprocessuser,
+    state_logprocessgroup0,
+    state_logprocessgroup,
     state_skip
   };
   
@@ -530,6 +718,9 @@ void log_init_cxxtools(std::istream& in)
   unsigned fsize;
   unsigned maxbackupindex;
   unsigned flushdelay = 0;
+  bool logprocess = false;
+  std::string logprocessuser;
+  std::string logprocessgroup;
 
   cxxtools::Logger::log_level_type level;
   while (in.get(ch))
@@ -569,8 +760,14 @@ void log_init_cxxtools(std::istream& in)
           state = state_maxbackupindex0;
         else if (ch == '=' && token == "FLUSHDELAY")
           state = state_flushdelay0;
-        else if (ch == '=' && token == "DISABLED")
+        else if (ch == '=' && (token == "DISABLE" || token == "DISABLED"))
           state = state_disable;
+        else if (ch == '=' && token == "LOGPROCESS")
+          state = state_logprocess;
+        else if (ch == '=' && token == "LOGPROCESSUSER")
+          state = state_logprocessuser0;
+        else if (ch == '=' && token == "LOGPROCESSGROUP")
+          state = state_logprocessgroup0;
         else if (ch == '\n')
           state = state_0;
         else if (std::isspace(ch))
@@ -597,8 +794,14 @@ void log_init_cxxtools(std::istream& in)
           state = state_maxbackupindex0;
         else if (ch == '=' && token == "FLUSHDELAY")
           state = state_flushdelay0;
-        else if (ch == '=' && token == "DISABLE")
+        else if (ch == '=' && (token == "DISABLE" || token == "DISABLED"))
           state = state_disable;
+        else if (ch == '=' && token == "LOGPROCESS")
+          state = state_logprocess;
+        else if (ch == '=' && token == "LOGPROCESSUSER")
+          state = state_logprocessuser0;
+        else if (ch == '=' && token == "LOGPROCESSGROUP")
+          state = state_logprocessgroup0;
         else if (ch == '\n')
           state = state_0;
         else if (!std::isspace(ch))
@@ -625,12 +828,18 @@ void log_init_cxxtools(std::istream& in)
 
         switch (ch)
         {
-          case 'F': level = cxxtools::Logger::LOG_LEVEL_FATAL; break;
-          case 'E': level = cxxtools::Logger::LOG_LEVEL_ERROR; break;
-          case 'W': level = cxxtools::Logger::LOG_LEVEL_WARN; break;
-          case 'I': level = cxxtools::Logger::LOG_LEVEL_INFO; break;
-          case 'D': level = cxxtools::Logger::LOG_LEVEL_DEBUG; break;
-          case 'T': level = cxxtools::Logger::LOG_LEVEL_TRACE; break;
+          case 'F':
+          case 'f': level = cxxtools::Logger::LOG_LEVEL_FATAL; break;
+          case 'E':
+          case 'e': level = cxxtools::Logger::LOG_LEVEL_ERROR; break;
+          case 'W':
+          case 'w': level = cxxtools::Logger::LOG_LEVEL_WARN; break;
+          case 'I':
+          case 'i': level = cxxtools::Logger::LOG_LEVEL_INFO; break;
+          case 'D':
+          case 'd': level = cxxtools::Logger::LOG_LEVEL_DEBUG; break;
+          case 'T':
+          case 't': level = cxxtools::Logger::LOG_LEVEL_TRACE; break;
           default:  level = cxxtools::Logger::getStdLevel(); break;
         }
         if (state == state_rootlevel)
@@ -779,6 +988,54 @@ void log_init_cxxtools(std::istream& in)
           state = (ch == '\n' ? state_0 : state_skip);
         break;
 
+      case state_logprocess:
+        if (ch == '1' || ch == 't' || ch == 'T' || ch == 'y' || ch == 'Y')
+        {
+          logprocess = true;
+          state = state_skip;
+        }
+        else if (ch != ' ' && ch != '\t')
+          state = (ch == '\n' ? state_0 : state_skip);
+        break;
+
+      case state_logprocessuser0:
+        if (ch == '\n')
+          state = state_0;
+        else if (!std::isspace(ch))
+        {
+          logprocessuser = ch;
+          state = state_logprocessuser;
+        }
+        break;
+
+      case state_logprocessuser:
+        if (ch == '\n')
+          state = state_0;
+        else if (std::isspace(ch))
+          state = state_skip;
+        else
+          logprocessuser += ch;
+        break;
+
+      case state_logprocessgroup0:
+        if (ch == '\n')
+          state = state_0;
+        else if (!std::isspace(ch))
+        {
+          logprocessgroup = ch;
+          state = state_logprocessgroup;
+        }
+        break;
+
+      case state_logprocessgroup:
+        if (ch == '\n')
+          state = state_0;
+        else if (std::isspace(ch))
+          state = state_skip;
+        else
+          logprocessgroup += ch;
+        break;
+
       case state_skip:
         if (ch == '\n')
           state = state_0;
@@ -786,6 +1043,9 @@ void log_init_cxxtools(std::istream& in)
     }
   }
 
+  if (logprocess)
+    cxxtools::LoggerImpl::runLoggerProcess(logprocessuser, logprocessgroup);
+
   cxxtools::LoggerImpl::setFlushDelay(flushdelay);
   cxxtools::reinitializeLoggers();
 }
