@@ -25,9 +25,11 @@
 #include <errno.h>
 #include <sys/poll.h>
 #include <sstream>
+#include <stdio.h>
+#include <string.h>
 #include "cxxtools/log.h"
 
-log_define("cxxtools.net");
+log_define("cxxtools.net")
 
 namespace cxxtools
 {
