@@ -25,7 +25,7 @@
 #include <sstream>
 #include <string>
 
-log_define("cxxtools.thread");
+log_define("cxxtools.thread")
 
 namespace cxxtools
 {
@@ -69,6 +69,7 @@ void AttachedThread::create()
   int ret = pthread_create(&pthreadId, &pthread_attr, start, (void*)this);
   if (ret != 0)
     throw ThreadException(ret, "pthread_create");
+  joined = false;
 }
 
 void AttachedThread::join()
